﻿/// Copyright (c) 2011 Andrey Veselov. All rights reserved.
/// WebSite: http://andrey.moveax.ru 
/// Email: andrey@moveax.ru
/// This source is subject to the Microsoft Public License (Ms-PL).

namespace MVCDemo.Models.MetadataProviders
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.IO;
    using System.Linq;
    using System.Reflection;
    using System.Web.Mvc;
    using System.Xml.Linq;
    using Models.Extensions;

    public class XmlModelMetadataProvider : AssociatedMetadataProvider, IModelMetadataProvider
    {
        #region Private fields

        private readonly string[] _availableProperties = 
        {
            // string
            "DataTypeName",
            "Description",
            "DisplayFormatString",
            "DisplayName",
            "EditFormatString",
            "NullDisplayText",
            "ShortDisplayName",
            "SimpleDisplayText",
            "TemplateHint",
            "Watermark",

            // bool
            "ConvertEmptyStringToNull",
            "HideSurroundingHtml",
            "IsReadOnly",
            "IsRequired",
            "ShowForDisplay",
            "ShowForEdit",

            // IDictionary<string, object>
            "AdditionalValues"
        };

        private readonly string _metadataStorage;

        private string _currentModelTypeName = string.Empty;

        private XElement _xmlMetadata;

        private Type _defaultResourceType;

        #endregion

        public XmlModelMetadataProvider(string metadataStorage)
        {
            if (!Directory.Exists(metadataStorage)) {
                throw new ArgumentException(
                    string.Format("Invalid directory {0}", metadataStorage));
            }

            if (metadataStorage.EndsWith("\\")) {
                this._metadataStorage = metadataStorage;
            }
            else {  
                this._metadataStorage = string.Format("{0}\\", metadataStorage);
            }

            this.ValidateAvailablePropertiesList(); // called in debug mode only
        }

        #region IModelMetadataProvider Members

        public bool IsAbleToProvideMetadata(Type type)
        {
            if (type == null) {
                return false;
            }

            if (string.Equals(this._currentModelTypeName, type.FullName)) {
                return true;
            }

            string metadataFile = string.Format(
                "{0}{1}.config", this._metadataStorage, type.FullName);

            if (!File.Exists(metadataFile)) {
                return false;
            }

            this._xmlMetadata = XElement.Load(metadataFile);
            this._defaultResourceType = null;

            var resourceAttr = this._xmlMetadata.Attribute("resource");
            if (resourceAttr != null) {
                this._defaultResourceType = Type.GetType(resourceAttr.Value, throwOnError: true);
            }

            this._currentModelTypeName = type.FullName;

            return true;
        }

        #endregion

        #region AssociatedMetadataProvider method override

        protected override ModelMetadata CreateMetadata(
            IEnumerable<System.Attribute> attributes,
            Type containerType,
            Func<object> modelAccessor,
            Type modelType,
            string propertyName)
        {
            var metadata = new ModelMetadata(
                this, containerType, modelAccessor, modelType, propertyName);

            if (string.IsNullOrEmpty(propertyName)) {
                return metadata;
            }

            // Get the data for the specified property.
            var xmlPropertyElement = this._xmlMetadata.Elements("property").FirstOrDefault(
                element => string.Equals(element.Attribute("name").Value, propertyName));

            if (xmlPropertyElement == null) {
                return metadata;
            }

            // Set the metadata for the property using XML attributes.
            foreach (var element in xmlPropertyElement.Elements()) {
                if (!string.Equals(element.Name.LocalName, "metadata")) {
                    continue;
                }

                string xmlPropertyName = this.GetAndValidatePropertyName(element);

                if (string.Equals(xmlPropertyName, "AdditionalValues")) {
                    this.SetMetadataAdditionalValue(metadata, element);
                }
                else {
                    var metadataProperty = typeof(ModelMetadata).GetProperty(xmlPropertyName);
                    object value = Convert.ChangeType(
                        this.GetValueFromElement(element),
                        metadataProperty.PropertyType);

                    metadataProperty.SetValue(metadata, value, null);
                }
            }

            return metadata;
        }

        #endregion

        #region Private methods

        private string GetAndValidatePropertyName(XElement element)
        {
            string propertyName = element.GetValueOrNull("name");
            if (propertyName == null) {
                throw new InvalidDataException(
                    string.Format("The name was not set for the element '{0}'.", element));
            }

            // Validate the property name.
            var nameFound = this._availableProperties.
                FirstOrDefault(p => string.Equals(p, propertyName));

            if (nameFound == null) {
                throw new InvalidDataException(
                    string.Format("Unknown or forbidden property: {0}.", propertyName));
            }

            return propertyName;
        }

        private void SetMetadataAdditionalValue(ModelMetadata metadata, XElement element)
        {
            string key = element.GetValueOrNull("key");
            if (key == null) {
                throw new InvalidDataException(
                    string.Format("The key was not set for the element '{0}'.", element));
            }

            // Set the value's type from XML element (or use default System.String).
            string typeName = element.GetValueOrNull("type");
            if (typeName == null) {
                typeName = "System.String";
            }

            // Get AdditionalValues property and add the new value.
            var addValProperty = typeof(ModelMetadata).GetProperty("AdditionalValues");
            var addVal = (IDictionary<string, object>)addValProperty.GetValue(metadata, null);
            object value = Convert.ChangeType(
                this.GetValueFromElement(element),
                Type.GetType(typeName));

            addVal.Add(key, value);
        }

        private string GetValueFromElement(XElement element)
        {
            // Try to get the value.
            string value = element.GetValueOrNull("value");
            if (value != null) {
                return value;
            }

            // Try to get the value from the resource.
            string resourceName = element.GetValueOrNull("resname");
            if (resourceName == null) {
                throw new InvalidDataException(
                    string.Format("The value is not set for the attribute '{0}'.", element));
            }

            string resourceTypeName = element.GetValueOrNull("resource");
            Type resourceType = this._defaultResourceType;

            if (resourceTypeName != null) {
                resourceType = Type.GetType(resourceTypeName, throwOnError: true);
            }

            var resourceProperty = resourceType.GetProperty(
                resourceName,
                BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic);

            return (string)resourceProperty.GetValue(null, null);
        }

        #endregion

        #region Debug methods

        [Conditional("DEBUG")]
        private void ValidateAvailablePropertiesList()
        {
            PropertyInfo[] props = typeof(ModelMetadata).GetProperties();

            foreach (var propertyName in this._availableProperties) {
                if (props.FirstOrDefault(p => string.Equals(p.Name, propertyName)) == null) {
                    throw new InvalidDataException(
                        string.Format("Unknown ModelMetadata property: {0}.", propertyName));
                }
            }
        }

        #endregion
    }
}